@@ -29,6 +29,13 @@ cmd2.parsing.shlex_split = functools.partial(
 )
 
 
+def extract_paste(text):
+    for line in text.splitlines():
+        if line.startswith("#"):
+            continue
+        yield line.split()[0].lstrip()
+
+
 def ispath(text: str):
     """Return True if text starts with a LwM2M Path"""
     text = text.strip()
@@ -82,7 +89,7 @@ def valsplit(p: str):
     """Split value from path"""
     parts = p.split("=", maxsplit=1)
     if len(parts) == 1:
-        return parts[0], ""
+        return parts[0], None
     return parts[0], parts[1]
 
 
@@ -101,7 +108,7 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
     ):
         """Initialization of LwM2M CommandInterpreter"""
         shortcuts = dict(cmd2.DEFAULT_SHORTCUTS)
-        shortcuts.update({'fota': 'firmware_update', 'attr': 'write_attr'})
+        shortcuts.update({"fota": "firmware_update", "attr": "write_attr"})
         super().__init__(shortcuts=shortcuts, **kwargs)
         self.console = rich.console.Console()
         self.log = logging.getLogger("Intrpr")
@@ -116,7 +123,7 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         # Override and use rich console
         with self.console.capture() as capture:
             self.console.print(msg, end="", **kwargs)
-        super().poutput(capture.get().rstrip('\n'), end=end)
+        super().poutput(capture.get().rstrip("\n"), end=end)
 
     @functools.lru_cache(maxsize=20)
     def cache(self, endpoint: str) -> emqxlwm2m.lwm2m.Endpoint:
@@ -140,6 +147,12 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         """exit interpreter"""
         return True
 
+    def do_sleep(self, line):
+        """Sleep seconds"""
+        t = float(line.strip())
+        self.poutput(f"Sleeping for {t} seconds ...")
+        time.sleep(t)
+
     def do_hist(self, line):
         """Fzf select command from history and run it"""
         seen = {"q", "hist"}
@@ -155,6 +168,11 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
             self.poutput(cmd)
             self.runcmds_plus_hooks([cmd])
 
+    def do_paste(self, line):
+        text = cmd2.clipboard.get_paste_buffer()
+        for ep in extract_paste(text):
+            self.poutput(ep)
+
     @property
     def prompt(self):
         timestamp = format(dt.datetime.now(), "%H:%M:%S")
@@ -162,7 +180,7 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
             self.console.print(end="\n")
             self.console.print(f"[{timestamp}]", end=" ")
             self.console.print(self.engine.host, style="red", end=":")
-            self.console.print(f'[black]{self.engine.port}', end="")
+            self.console.print(f"[black]{self.engine.port}", end="")
             self.console.print(f"\n{self.ep_active}$", style="black", end=" ")
         return capture.get()
 
@@ -211,27 +229,23 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
 
         # If special word "paste", take endpoints from clipboard
         try:
-            args.endpoint.remove('paste')
+            args.endpoint.remove("paste")
         except ValueError:
             pass
         else:
             text = cmd2.clipboard.get_paste_buffer()
-            for line in text.splitlines():
-                if line.startswith('#'):
-                    continue
-                ep = line.strip()
-                if ep:
-                    args.endpoint.append(ep)
+            for ep in extract_paste(text):
+                args.endpoint.append(ep)
 
         # Make sure endpoint starts with ep_prefix
         for i in range(len(args.endpoint)):
             ep = args.endpoint[i]
             if not ep.startswith(self.ep_prefix):
-                args.endpoint[i] = f'{self.ep_prefix}{ep}'
+                args.endpoint[i] = f"{self.ep_prefix}{ep}"
 
         # Add endpoints to history file
         for ep in args.endpoint:
-            add_endpoint_to_history(endpoint)
+            add_endpoint_to_history(ep)
 
     def select_path(self):
         """Select path interactively with fzf"""
@@ -254,7 +268,7 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         if not args.path:
             raise Exception("Path argument missing")
         for p in args.path:
-            add_path_to_history(p.split('=')[0])
+            add_path_to_history(p.split("=")[0])
 
     def print_message(self, msg: emqxlwm2m.lwm2m.Message):
         """Print a LwM2M message in a pretty way"""
@@ -278,48 +292,48 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
 
         # Print endpoint name
         if isinstance(msg, emqxlwm2m.lwm2m.Downlink):
-            self.poutput(msg.ep, end='< ')
+            self.poutput(msg.ep, end=" < ")
         else:  # Uplink
-            self.poutput(msg.ep, end='> ')
+            self.poutput(msg.ep, end=" > ")
 
         # Print message type
         name = type(msg).__name__
         if isinstance(
             msg, (emqxlwm2m.lwm2m.Registration, emqxlwm2m.lwm2m.Update)
         ):
-            self.poutput(f'[bold black]{name}', end=' ')
+            self.poutput(f"[bold black]{name}", end=" ")
         elif isinstance(msg, emqxlwm2m.lwm2m.Notification):
-            self.poutput(f'[italic black]{name}', end=' ')
+            self.poutput(f"[italic black]{name}", end=" ")
         else:
-            self.poutput(name, end=' ')
+            self.poutput(name, end=" ")
 
         # Print registration or update
         if isinstance(
             msg, (emqxlwm2m.lwm2m.Registration, emqxlwm2m.lwm2m.Update)
         ):
-            self.poutput(f'lt={msg.lt}', end=' ')
-            self.poutput(f'sms={msg.sms}', end=' ')
-            self.poutput(f'lwm2m={msg.lwm2m}', end=' ')
-            self.poutput(f'b={msg.b}', end=' ')
-            self.poutput(f'alternate_path={msg.alternate_path}', end=' ')
+            self.poutput(f"lt={msg.lt}", end=" ")
+            self.poutput(f"sms={msg.sms}", end=" ")
+            self.poutput(f"lwm2m={msg.lwm2m}", end=" ")
+            self.poutput(f"b={msg.b}", end=" ")
+            self.poutput(f"alternate_path={msg.alternate_path}", end=" ")
             self.poutput(msg.data)
             return
 
         # Print requests
         if isinstance(msg, emqxlwm2m.lwm2m.Request):
             if timeout:
-                self.poutput('[bold red]TIMEOUT', end=' ')
+                self.poutput("[bold red]TIMEOUT", end=" ")
         if isinstance(msg, emqxlwm2m.lwm2m.WriteAttrRequest):
-            self.poutput(f'path={msg.path}', end=' ')
-            self.poutput(f'pmin={msg.pmin}', end=' ')
-            self.poutput(f'pmax={msg.pmax}', end=' ')
-            self.poutput(f'lt={msg.lt}', end=' ')
-            self.poutput(f'st={msg.st}', end=' ')
-            self.poutput(f'gt={msg.gt}')
+            self.poutput(f"path={msg.path}", end=" ")
+            self.poutput(f"pmin={msg.pmin}", end=" ")
+            self.poutput(f"pmax={msg.pmax}", end=" ")
+            self.poutput(f"lt={msg.lt}", end=" ")
+            self.poutput(f"st={msg.st}", end=" ")
+            self.poutput(f"gt={msg.gt}")
             return
         if isinstance(msg, emqxlwm2m.lwm2m.ExecuteRequest):
-            self.poutput(f'path={msg.path}', end=' ')
-            self.poutput(f'args={msg.args!r}')
+            self.poutput(f"path={msg.path}", end=" ")
+            self.poutput(f"args={msg.args!r}")
             return
         if isinstance(
             msg, (emqxlwm2m.lwm2m.WriteRequest, emqxlwm2m.lwm2m.CreateRequest)
@@ -327,21 +341,21 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
             self.poutput(msg.data)
             return
         if isinstance(msg, emqxlwm2m.lwm2m.Request):
-            self.poutput(f'path={msg.path}')
+            self.poutput(f"path={msg.path}")
             return
 
         # Handle responses
         if error:
-            color = '[bold red]'
+            color = "[bold red]"
         else:
-            color = '[green]'
-        self.poutput(f'{color}{msg.code.name} {msg.code.value}[/]', end=' ')
+            color = "[green]"
+        self.poutput(f"{color}{msg.code.name} {msg.code.value}[/]", end=" ")
         if isinstance(msg, emqxlwm2m.lwm2m.Notification):
-            self.poutput(f'Seq={msg.seq_num}', end=' ')
-        if hasattr(msg, 'data'):
+            self.poutput(f"Seq={msg.seq_num}", end=" ")
+        if hasattr(msg, "data"):
             self.poutput(msg.data)
         else:
-            self.poutput(f'{color}ReqPath {msg.req_path}[/]')
+            self.poutput(f"{color}ReqPath {msg.req_path}[/]")
 
     def print_notifications(self, notify_queue, stop_after=10 ** 10):
         tracker = emqxlwm2m.lwm2m.NotificationsTracker()
@@ -408,14 +422,11 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         pathvals = list()
         for path in args.path:
             p, value = valsplit(path)
-            if not value:
+            if value is None:
                 if args.value is not None:
                     value = args.value
                 else:
                     value = input(f"Write value to {p}: ")
-                if not value:
-                    print("No value provided! Aborting.")
-                    return
             value = emqxlwm2m.engines.emqx.str_to_py_value(value)
             pathvals.append((Path(p), value))
         pathvals.sort()
@@ -433,11 +444,10 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
 
                     groups = itertools.groupby(pathvals, key=group)
                     for _, chunk in groups:
-                        values = dict()
-                        for p, value in chunk:
-                            values[p] = value
                         try:
-                            resp = ep.write(None, values, timeout=args.timeout)
+                            resp = ep.write(
+                                "", dict(chunk), timeout=args.timeout
+                            )
                             self.print_message(resp)
                         except emqxlwm2m.ResponseError as error:
                             self.print_message(error)
@@ -505,7 +515,7 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
                 break
             time.sleep(args.repeat)
 
-    def complete_attr(self, text, line, begidx, endidx):
+    def complete_write_attr(self, text, line, begidx, endidx):
         return self._complete_path(text, line, begidx, endidx)
 
     # -------------------------------------------------------------------------
@@ -518,7 +528,7 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         pathargs = list()
         for path in args.path:
             p, arg = valsplit(path)
-            if not arg:
+            if arg is None:
                 if args.arg is not None:
                     arg = args.arg
                 else:
@@ -551,7 +561,7 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         pathvals = list()
         for path in args.path:
             p, value = valsplit(path)
-            if not value:
+            if value is None:
                 if args.value is not None:
                     value = args.value
                 else:
@@ -573,11 +583,8 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
 
                 groups = itertools.groupby(pathvals, key=group)
                 for _, chunk in groups:
-                    values = dict()
-                    for p, value in chunk:
-                        values[p] = value
                     try:
-                        resp = ep.create(values, timeout=args.timeout)
+                        resp = ep.create("", dict(chunk), timeout=args.timeout)
                         self.print_message(resp)
                     except emqxlwm2m.ResponseError as error:
                         self.print_message(error)
@@ -622,9 +629,12 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
                 for path in args.path:
                     path, _ = valsplit(path)
                     try:
-                        resp, q = ep.observe(
-                            path, queue=q, timeout=args.timeout
+                        resp = ep.observe(
+                            path,
+                            timeout=args.timeout,
+                            queue=q,
                         )
+                        q = resp.notifications
                         self.print_message(resp)
                     except emqxlwm2m.ResponseError as error:
                         self.print_message(error)
@@ -693,6 +703,16 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         for _ in range(args.stop_after):
             self.print_message(q.get())
 
+    @cmd2.with_argparser(parser.commands)
+    @endpoint
+    def do_commands(self, args):
+        q = None
+        for ep in args.endpoint:
+            ep = self.cache(ep)
+            q = ep.commands(queue=q)
+        for _ in range(args.stop_after):
+            self.print_message(q.get())
+
     @cmd2.with_argparser(parser.requests)
     @endpoint
     def do_requests(self, args):
@@ -762,7 +782,28 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         for _ in range(args.count):
             for ep in args.endpoint:
                 ep = self.cache(ep)
-                emqxlwm2m.utils.reboot(ep, args.wait, args.timeout)
+                t0 = dt.datetime.now()
+                if args.wait:
+                    q = ep.registrations()
+                self.poutput(f"{ep.endpoint}: Rebooting")
+                try:
+                    resp = ep.execute("3/0/4", timeout=args.timeout)
+                    self.print_message(resp)
+                except emqxlwm2m.ResponseError as error:
+                    self.print_message(error)
+                else:
+                    if args.wait:
+                        self.poutput(
+                            f"{ep.endpoint}: Waiting for registration"
+                        )
+                        reg = q.get()
+                        self.print_message(reg)
+                        t1 = reg.timestamp
+                        self.poutput(
+                            f"{ep.endpoint}: Reboot duration {t1 - t0}",
+                        )
+                if len(args.endpoint):
+                    time.sleep(args.pause)
             if args.repeat is None:
                 break
             time.sleep(args.repeat)
@@ -806,7 +847,14 @@ class CommandInterpreter(cmd2.Cmd):  # Todo: plugins
         for ep in args.endpoint:
             ep = self.cache(ep)
             for _ in range(args.retry + 1):
-                if emqxlwm2m.utils.firmware_update(ep, args.package_uri):
-                    if args.reboot:
-                        emqxlwm2m.utils.reboot(ep, True, args.timeout)
-                    break
+                try:
+                    if emqxlwm2m.utils.firmware_update(
+                        ep, args.package_uri, reg_timeout=1_000
+                    ):
+                        if args.reboot:
+                            emqxlwm2m.utils.reboot(ep, True, args.timeout)
+                        break
+                except emqxlwm2m.ResponseError as error:
+                    self.print_message(error)
+                except emqxlwm2m.utils.FirmwareUpdateError as error:
+                    ep.log.error("%s: %s", type(error).__name__, error)
