@@ -1,44 +1,62 @@
-
 # Built-in
+import asyncio
 import collections
+import contextlib
 import dataclasses
 import datetime as dt
 import enum
 import functools
+import logging
 import re
 import typing
 
 
 class ResponseError(Exception):
-    pass
+    """Timeout or CoAP response code error"""
 
 
 class NoResponseError(ResponseError):
-    pass
+    """LwM2M operation timeout"""
 
 
 class CoAPResponseCode(enum.Enum):
-    Created = '2.01'
-    Deleted = '2.02'
-    Valid = '2.03'
-    Changed = '2.04'
-    Content = '2.05'
-    BadRequest = '4.00'
-    Unauthorized = '4.01'
-    BadOption = '4.02'
-    Forbidden = '4.03'
-    NotFound = '4.04'
-    MethodNotAllowed = '4.05'
-    NotAcceptable = '4.06'
-    PreconditionFailed = '4.12'
-    RequestEntityTooLarge = '4.13'
-    UnsupportedContentFormat = '4.15'
-    InternalServerError = '5.00'
-    NotImplemented_ = '5.01'
-    BadGateway = '5.02'
-    ServiceUnavailable = '5.03'
-    GatewayTimeout = '5.04'
-    ProxyingNotSupported = '5.05'
+    """The Code field in the CoAP header being set to a Response Code.
+
+    Similar to the HTTP Status Code, the CoAP Response Code indicates
+    the result of the attempt to understand and satisfy the request.
+    """
+
+    # Success 2.xx. This class of Response Code indicates that the
+    # clients request was successfully received, understood, and
+    # accepted.
+    Created = "2.01"
+    Deleted = "2.02"
+    Valid = "2.03"
+    Changed = "2.04"
+    Content = "2.05"
+    # Client Error 4.xx. This class of Response Code is intended for
+    # cases in which the client seems to have erred.  These Response
+    # Codes are applicable to any request method.
+    BadRequest = "4.00"
+    Unauthorized = "4.01"
+    BadOption = "4.02"
+    Forbidden = "4.03"
+    NotFound = "4.04"
+    MethodNotAllowed = "4.05"
+    NotAcceptable = "4.06"
+    PreconditionFailed = "4.12"
+    RequestEntityTooLarge = "4.13"
+    UnsupportedContentFormat = "4.15"
+    # Server Error 5.xx. This class of Response Code indicates cases
+    # in which the server is aware that it has erred or is incapable
+    # of performing the request.  These Response Codes are applicable
+    # to any request method.
+    InternalServerError = "5.00"
+    NotImplemented_ = "5.01"
+    BadGateway = "5.02"
+    ServiceUnavailable = "5.03"
+    GatewayTimeout = "5.04"
+    ProxyingNotSupported = "5.05"
 
 
 class BadPath(Exception):
@@ -47,16 +65,11 @@ class BadPath(Exception):
 
 class Path(collections.UserString):
 
-    levels = [
-        'object',
-        'object_instance',
-        'resource',
-        'resource_instance'
-    ]
+    levels = ["object", "object_instance", "resource", "resource_instance"]
 
     def __init__(self, seq):
         """Initialization of Path"""
-        seq = re.sub('/+', '/', str(seq))
+        seq = re.sub("/+", "/", str(seq))
         super().__init__(seq)
 
     @classmethod
@@ -65,50 +78,50 @@ class Path(collections.UserString):
 
     @property
     def parts(self):
-        return [p for p in self.data.lstrip('/').split('/') if p]
+        return [p for p in self.data.lstrip("/").split("/") if p]
 
     @property
     def level(self):
         n = len(self.parts) - 1
         if n < 0:
-            return 'root'
+            return "root"
         return self.levels[n]
 
     @property
     def oid(self):
         try:
-            return int(self.parts[self.levels.index('object')])
+            return int(self.parts[self.levels.index("object")])
         except IndexError as error:
-            raise BadPath('No object id!') from error
+            raise BadPath("No object id!") from error
         except (ValueError, TypeError) as error:
-            raise BadPath('Object id is not an integer') from error
+            raise BadPath("Object id is not an integer") from error
 
     @property
     def iid(self):
         try:
-            return int(self.parts[self.levels.index('object_instance')])
+            return int(self.parts[self.levels.index("object_instance")])
         except IndexError as error:
-            raise BadPath('No object instance id!') from error
+            raise BadPath("No object instance id!") from error
         except (ValueError, TypeError) as error:
-            raise BadPath('Instance id is not an integer') from error
+            raise BadPath("Instance id is not an integer") from error
 
     @property
     def rid(self):
         try:
-            return int(self.parts[self.levels.index('resource')])
+            return int(self.parts[self.levels.index("resource")])
         except IndexError as error:
-            raise BadPath('No resource id!') from error
+            raise BadPath("No resource id!") from error
         except (ValueError, TypeError) as error:
-            raise BadPath('Resource id is not an integer') from error
+            raise BadPath("Resource id is not an integer") from error
 
     @property
     def riid(self):
         try:
-            return int(self.parts[self.levels.index('resource_instance')])
+            return int(self.parts[self.levels.index("resource_instance")])
         except IndexError as error:
-            raise BadPath('No resource instance id!') from error
+            raise BadPath("No resource instance id!") from error
         except (ValueError, TypeError) as error:
-            raise BadPath('Resource instance id is not an integer') from error
+            raise BadPath("Resource instance id is not an integer") from error
 
 
 class NotificationsTracker:
@@ -142,7 +155,7 @@ class NotificationsTracker:
         """Return list of duration between packets"""
         history = list(self._history[packet.ep, packet.req_path])
         if not history:
-            return [float('inf')]
+            return [float("inf")]
         try:
             history.remove(packet)
         except ValueError:
@@ -153,9 +166,11 @@ class NotificationsTracker:
             output.append((p2.timestamp - p1.timestamp).total_seconds())
         return output
 
+
 @dataclasses.dataclass
 class Attributes:
     """Container for LwM2M attributes"""
+
     pmin: int = None
     pmax: int = None
     lt: float = None
@@ -163,39 +178,40 @@ class Attributes:
     gt: float = None
 
     regex = re.compile(
-        r'(\[(?P<pmin>\d*),\s*(?P<pmax>\d*)\])?'\
-        r'(\s*(?P<lt>[^:]*?):(?P<st>[^:]*?):(?P<gt>[^:]*))?')
+        r"(\[(?P<pmin>\d*),\s*(?P<pmax>\d*)\])?"
+        r"(\s*(?P<lt>[^:]*?):(?P<st>[^:]*?):(?P<gt>[^:]*))?"
+    )
 
     @classmethod
     def from_string(cls, string):
         """Format: [pmin,pmax]lt:st:gt"""
         match = cls.regex.match(string)
         if not match:
-            raise ValueError(f'Bad format: {string!r}')
+            raise ValueError(f"Bad format: {string!r}")
         # Replace empyt string with None
         d = match.groupdict()
         for k, v in d.items():
-            if v == '':
+            if v == "":
                 d[k] = None
-        for attr in ('pmin', 'pmax'):
+        for attr in ("pmin", "pmax"):
             if d.get(attr) is not None:
                 d[attr] = int(d[attr])
-        for attr in ('lt', 'st', 'gt'):
+        for attr in ("lt", "st", "gt"):
             if d.get(attr) is not None:
                 d[attr] = float(d[attr])
         return cls(**d)
 
     def __str__(self):
-        output = ''
-        pmin = self.pmin or ''
-        pmax = self.pmax or ''
-        lt = self.lt or ''
-        st = self.st or ''
-        gt = self.gt or ''
+        output = ""
+        pmin = self.pmin or ""
+        pmax = self.pmax or ""
+        lt = self.lt or ""
+        st = self.st or ""
+        gt = self.gt or ""
         if any((pmin, pmax)):
-            output += f'[{pmin},{pmax}]'
+            output += f"[{pmin},{pmax}]"
         if any((lt, st, gt)):
-            output += f'{lt}:{st}:{gt}'
+            output += f"{lt}:{st}:{gt}"
         return output
 
     def __iter__(self):
@@ -203,7 +219,10 @@ class Attributes:
 
     def __len__(self):
         """Return number of attributes not None"""
-        return len([a for a in dataclasses.asdict(self).values() if a is not None])
+        return len(
+            [a for a in dataclasses.asdict(self).values() if a is not None]
+        )
+
 
 # ========
 # PAYLOADS
@@ -215,7 +234,8 @@ from dataclasses import dataclass, field
 class Message:
     ep: str
     timestamp: dt.datetime = field(
-        init=False, repr=False, default_factory=dt.datetime.now, compare=False)
+        init=False, repr=False, default_factory=dt.datetime.now, compare=False
+    )
 
 
 class Downlink(Message):
@@ -223,7 +243,6 @@ class Downlink(Message):
 
 
 class Request(Downlink):
-
     def __post_init__(self):
         try:
             self.path = Path(self.path)
@@ -314,20 +333,19 @@ class Response(Uplink):
             pass
 
     def check(self):
-        if not self.code.value.startswith('2.'):
+        if not self.code.value.startswith("2."):
             raise ResponseError(self)
 
     @property
     def value(self):
+        self.check()
         try:
-            self.check()
             return self.data[self.req_path]
         except AttributeError:
             raise AttributeError(
-                f'Responses of type {type(self)} has no value') \
-                from None
+                f"Responses of type {type(self)} has no value"
+            ) from None
         except KeyError:
-            self.check()
             return self.data
 
 
@@ -447,27 +465,64 @@ class Notification(Event, collections.UserDict):
         self.data = Path.dict(self.data)
 
     def check(self):
-        if not self.code.value.startswith('2.'):
+        if not self.code.value.startswith("2."):
             raise ResponseError(self)
 
+    @property
+    def value(self):
+        self.check()
+        try:
+            return self.data[self.req_path]
+        except KeyError:
+            return self.data
+
+
 # ==============
 # LWM2M ENDPOINT
 # ==============
 
+
 class Endpoint:
     """LwM2M for specific endpoint"""
 
-    def __init__(self, endpoint, timeout=None):
+    def __init__(self, endpoint: str, timeout: float = None):
         self.endpoint = endpoint
         self.timeout = timeout
         self.engine = None
+        self._log = None
 
     def __repr__(self):
-        return f'{self.__class__}({self.endpoint})'
+        return f"{self.__class__}({self.endpoint})"
 
     def __str__(self):
         return self.endpoint
 
+    @property
+    def log(self):
+        # Create only if needed
+        if self._log is None:
+            self._log = logging.getLogger(self.endpoint)
+        return self._log
+
+    def _send(self, msg: Message, timeout: float, retry: int = 0) -> Message:
+        """Call engine.send() with default timeout and retry logic"""
+        if timeout is None:
+            timeout = self.timeout
+        for i in range(retry):
+            try:
+                return self.engine.send(msg, timeout)
+            except NoResponseError as error:
+                # Log only if logger already exist
+                if self._log is not None:
+                    self.log.warning(
+                        "%s (retry %d/%d): %s",
+                        type(error).__name__,
+                        i + 1,
+                        retry,
+                        str(error),
+                    )
+        return self.engine.send(msg, timeout)
+
     def wiretap(self, *, queue=None):
         return self.engine.recv(self.endpoint, [Message], queue=queue)
 
@@ -506,74 +561,86 @@ class Endpoint:
         msgs = [Request, Response]
         return self.engine.recv(self.endpoint, msgs, queue=queue)
 
-    def discover(self, path, timeout=None) -> DiscoverResponse:
-        if timeout is None:
-            timeout = self.timeout
+    def discover(
+        self, path, timeout: float = None, retry: int = 0
+    ) -> DiscoverResponse:
         msg = DiscoverRequest(self.endpoint, path)
-        return self.engine.send(msg, timeout)
+        return self._send(msg, timeout, retry)
 
-    def read(self, path, timeout=None) -> ReadResponse:
-        if timeout is None:
-            timeout = self.timeout
+    def read(
+        self, path, timeout: float = None, retry: int = 0
+    ) -> ReadResponse:
         msg = ReadRequest(self.endpoint, path)
-        return self.engine.send(msg, timeout)
+        return self._send(msg, timeout, retry)
 
-    def write(self, path, value, timeout=None) -> WriteResponse:
-        if timeout is None:
-            timeout = self.timeout
+    def write(
+        self, path, value, timeout: float = None, retry: int = 0
+    ) -> WriteResponse:
         if isinstance(value, dict):
-            data = value
+            if path:
+                data = {path + "/" + p: v for p, v in value.items()}
+            else:
+                data = value
         else:
             data = {path: value}
         msg = WriteRequest(self.endpoint, data)
-        return self.engine.send(msg, timeout)
+        return self._send(msg, timeout, retry)
 
-    def write_attr(self, path, pmin=None, pmax=None,
-                   lt=None, st=None, gt=None,
-                   timeout=None) -> WriteAttrResponse:
-        if timeout is None:
-            timeout = self.timeout
+    def write_attr(
+        self,
+        path,
+        pmin: int = None,
+        pmax: int = None,
+        lt: float = None,
+        st: float = None,
+        gt: float = None,
+        timeout: float = None,
+        retry: int = 0,
+    ) -> WriteAttrResponse:
         msg = WriteAttrRequest(self.endpoint, path, pmin, pmax, lt, st, gt)
-        return self.engine.send(msg, timeout)
+        return self._send(msg, timeout, retry)
 
-    def execute(self, path, args="", timeout=None) -> ExecuteResponse:
-        if timeout is None:
-            timeout = self.timeout
+    def execute(
+        self, path, args="", timeout: float = None, retry: int = 0
+    ) -> ExecuteResponse:
         msg = ExecuteRequest(self.endpoint, path, args)
-        return self.engine.send(msg, timeout)
+        return self._send(msg, timeout, retry)
 
-    def create(self, values, timeout=None) -> CreateResponse:
-        """Create object instance
+    def create(
+        self, path, value, timeout: float = None, retry: int = 0
+    ) -> CreateResponse:
+        if isinstance(value, dict):
+            if path:
+                data = {path + "/" + p: v for p, v in value.items()}
+            else:
+                data = value
+        else:
+            data = {path: value}
+        msg = CreateRequest(self.endpoint, data)
+        return self._send(msg, timeout, retry)
 
-        Content needed for all mandatory resources. E.g:
-
-        values = {'123/1/0': 'test', '123/1/1': 3.1415}
-        """
-        if timeout is None:
-            timeout = self.timeout
-        msg = CreateRequest(self.endpoint, values)
-        return self.engine.send(msg, timeout)
-
-    def delete(self, path, timeout=None) -> DeleteResponse:
-        if timeout is None:
-            timeout = self.timeout
+    def delete(
+        self, path, timeout: float = None, retry: int = 0
+    ) -> DeleteResponse:
         msg = DeleteRequest(self.endpoint, path)
-        return self.engine.send(msg, timeout)
+        return self._send(msg, timeout, retry)
 
     # LwM2M Information Reporting Interface
     # -------------------------------------
 
-    def observe(self, path, timeout=None, *, queue=None) -> ObserveResponse:
-        if timeout is None:
-            timeout = self.timeout
+    def observe(
+        self, path, timeout: float = None, retry: int = 0, *, queue=None
+    ) -> ObserveResponse:
         msg = ObserveRequest(self.endpoint, path)
-        return self.engine.send(msg, timeout)
+        if queue:
+            msg.queue = queue
+        return self._send(msg, timeout, retry)
 
-    def cancel_observe(self, path, timeout=None) -> CancelObserveResponse:
-        if timeout is None:
-            timeout = self.timeout
+    def cancel_observe(
+        self, path, timeout: float = None, retry: int = 0
+    ) -> CancelObserveResponse:
         msg = CancelObserveRequest(self.endpoint, path)
-        return self.engine.send(msg, timeout)
+        return self._send(msg, timeout, retry)
 
     def notifications(self, *, queue=None):
         return self.engine.recv(self.endpoint, [Notification], queue=queue)
@@ -598,86 +665,255 @@ class Endpoint:
         """
         if issubclass(object_def, ObjectDef):
             return object_def(self)
-        raise TypeError('Key must be subclass of ObjectDef')
+        raise TypeError("Key must be subclass of ObjectDef")
+
+    # Context managers
+    # ----------------
+
+    @contextlib.contextmanager
+    def temporary_observe(
+        self, path, timeout: float = None, retry: int = 0, *, queue=None
+    ):
+        """Observe at enter, cancel observe at exit"""
+        try:
+            yield self.observe(path, timeout, retry, queue=queue)
+        finally:
+            self.cancel_observe(path, timeout, retry)
 
 
-def use_enums(method):
-    """Replace values with enums when possible"""
-    @functools.wraps(method)
-    def wrapper(self, *args, **kwargs):
-        resp = method(self, *args, **kwargs)
-        for p, v in resp.items():
-            try:
-                rid = p.rid
-            except BadPath:
-                continue
-            else:
-                if isinstance(self, Operation):
-                    r = self.resource
-                else:
-                    r = self.resource_by_id(rid)
-                    if r is None:
-                        continue
-                try:
-                    if issubclass(r.type, enum.Enum):
-                        resp[p] = r.type(v)
-                except TypeError:
-                    continue
-        return resp
-    return wrapper
+class AsyncEndpoint:
+    """LwM2M for specific endpoint"""
 
+    def __init__(self, endpoint: str, timeout: float = None):
+        self.endpoint = endpoint
+        self.timeout = timeout
+        self.engine = None
+        self._log = None
 
-class LwM2MPath:
+    def __repr__(self):
+        return f"{self.__class__}({self.endpoint})"
 
-    def __init__(self, endpoint: Endpoint):
-        self.ep = endpoint
+    def __str__(self):
+        return self.endpoint
 
     @property
-    def path(self) -> Path:
-        raise NotImplementedError
+    def log(self):
+        # Create only if needed
+        if self._log is None:
+            self._log = logging.getLogger(self.endpoint)
+        return self._log
 
-    def discover(self, timeout=None):
-        return self.ep.discover(self.path, timeout)
+    async def _send(
+        self, msg: Message, timeout: float, retry: int = 0
+    ) -> Message:
+        """Call engine.send() with default timeout and retry logic"""
+        if timeout is None:
+            timeout = self.timeout
 
-    @use_enums
-    def read(self, timeout=None):
-        return self.ep.read(self.path, timeout)
+        for i in range(retry):
+            try:
+                return await self.engine.send(msg, timeout)
+            except NoResponseError as error:
+                # Log only if logger already exist
+                if self._log is not None:
+                    self.log.warning(
+                        "%s (retry %d/%d): %s",
+                        type(error).__name__,
+                        i + 1,
+                        retry,
+                        str(error),
+                    )
+        return await self.engine.send(msg, timeout)
 
-    def write(self, value, timeout=None):
-        if isinstance(value, dict):  # Batch write
-            values = dict()
-            for rid, v in value.items():
-                values[f'/{self.path.oid}/{self.path.iid}/{rid}'] = v
-            return self.ep.write(None, values, timeout)
-        return self.ep.write(self.path, value, timeout)
+    async def wiretap(self, *, queue=None):
+        return await self.engine.recv(self.endpoint, [Message], queue=queue)
 
-    def write_attr(self, pmin=None, pmax=None,
-                   lt=None, st=None, gt=None, timeout=None):
-        return self.ep.write_attr(self.path, pmin, pmax,
-                                  lt, st, gt, timeout)
+    async def uplink(self, *, queue=None):
+        return await self.engine.recv(self.endpoint, [Uplink], queue=queue)
 
-    def execute(self, args="", timeout=None):
-        return self.ep.execute(self.path, args, timeout)
+    async def downlink(self, *, queue=None):
+        return await self.engine.recv(self.endpoint, [Downlink], queue=queue)
 
-    def create(self, values, timeout=None):
-        data = dict()
-        for rid, v in values.items():
-            data[f'/{self.path.oid}/{self.path.iid}/{rid}'] = v
-        return self.ep.create(data, timeout)
+    async def requests(self, *, queue=None):
+        return await self.engine.recv(self.endpoint, [Request], queue=queue)
 
-    def delete(self, timeout=None):
-        return self.ep.delete(self.path, timeout)
+    async def responses(self, *, queue=None):
+        return await self.engine.recv(self.endpoint, [Response], queue=queue)
+
+    async def events(self, *, queue=None):
+        return await self.engine.recv(self.endpoint, [Event], queue=queue)
+
+    # LwM2M Client Registration Interface
+    # -----------------------------------
+
+    async def registrations(self, *, include_updates=False, queue=None):
+        msgs = [Registration]
+        if include_updates:
+            msgs.append(Update)
+        return await self.engine.recv(self.endpoint, msgs, queue=queue)
+
+    async def updates(self, *, queue=None):
+        msgs = [Update]
+        return await self.engine.recv(self.endpoint, msgs, queue=queue)
+
+    # LwM2M Device Management & Service Enablement Interface
+    # ------------------------------------------------------
+
+    async def commands(self, *, queue=None):
+        msgs = [Request, Response]
+        return await self.engine.recv(self.endpoint, msgs, queue=queue)
+
+    async def discover(
+        self, path, timeout: float = None, retry: int = 0
+    ) -> DiscoverResponse:
+        msg = DiscoverRequest(self.endpoint, path)
+        return await self._send(msg, timeout, retry)
+
+    async def read(
+        self, path, timeout: float = None, retry: int = 0
+    ) -> ReadResponse:
+        msg = ReadRequest(self.endpoint, path)
+        return await self._send(msg, timeout, retry)
+
+    async def write(
+        self, path, value, timeout: float = None, retry: int = 0
+    ) -> WriteResponse:
+        if isinstance(value, dict):
+            if path:
+                data = {path + "/" + p: v for p, v in value.items()}
+            else:
+                data = value
+        else:
+            data = {path: value}
+        msg = WriteRequest(self.endpoint, data)
+        return await self._send(msg, timeout, retry)
+
+    async def write_attr(
+        self,
+        path,
+        pmin: int = None,
+        pmax: int = None,
+        lt: float = None,
+        st: float = None,
+        gt: float = None,
+        timeout: float = None,
+        retry: int = 0,
+    ) -> WriteAttrResponse:
+        msg = WriteAttrRequest(self.endpoint, path, pmin, pmax, lt, st, gt)
+        return await self._send(msg, timeout, retry)
+
+    async def execute(
+        self, path, args="", timeout: float = None, retry: int = 0
+    ) -> ExecuteResponse:
+        msg = ExecuteRequest(self.endpoint, path, args)
+        return await self._send(msg, timeout, retry)
+
+    async def create(
+        self, path, value, timeout: float = None, retry: int = 0
+    ) -> CreateResponse:
+        if isinstance(value, dict):
+            if path:
+                data = {path + "/" + p: v for p, v in value.items()}
+            else:
+                data = value
+        else:
+            data = {path: value}
+        msg = CreateRequest(self.endpoint, data)
+        return await self._send(msg, timeout, retry)
+
+    async def delete(
+        self, path, timeout: float = None, retry: int = 0
+    ) -> DeleteResponse:
+        msg = DeleteRequest(self.endpoint, path)
+        return await self._send(msg, timeout, retry)
 
     # LwM2M Information Reporting Interface
     # -------------------------------------
 
-    @use_enums
-    def observe(self, timeout=None, *, queue=None):
-        return self.ep.observe(self.path, timeout=timeout, queue=queue)
+    async def observe(
+        self, path, timeout: float = None, retry: int = 0, *, queue=None
+    ) -> ObserveResponse:
+        msg = ObserveRequest(self.endpoint, path)
+        if queue:
+            msg.queue = queue
+        return await self._send(msg, timeout, retry)
 
-    @use_enums
-    def cancel_observe(self, timeout=None):
-        return self.ep.cancel_observe(self.path, timeout)
+    async def cancel_observe(
+        self, path, timeout: float = None, retry: int = 0
+    ) -> CancelObserveResponse:
+        msg = CancelObserveRequest(self.endpoint, path)
+        return await self._send(msg, timeout, retry)
+
+    async def notifications(self, *, queue=None):
+        return await self.engine.recv(
+            self.endpoint, [Notification], queue=queue
+        )
+
+    # LwM2M Object generated paths
+    # ----------------------------
+
+    def __getitem__(self, object_def):
+        if issubclass(object_def, ObjectDef):
+            return object_def(self)
+        raise TypeError("Key must be subclass of ObjectDef")
+
+    # Context managers
+    # ----------------
+
+    @contextlib.asynccontextmanager
+    async def temporary_observe(
+        self, path, timeout: float = None, retry: int = 0, *, queue=None
+    ):
+        """Observe at enter, cancel observe at exit"""
+        try:
+            yield await self.observe(path, timeout, retry, queue=queue)
+        finally:
+            await self.cancel_observe(path, timeout, retry)
+
+
+def replace_with_enums(self, resp):
+    for p, v in resp.items():
+        try:
+            rid = p.rid
+        except BadPath:
+            continue
+        else:
+            if isinstance(self, Operation):
+                r = self.resource
+            else:
+                r = self.resource_by_id(rid)
+                if r is None:
+                    continue
+            try:
+                if issubclass(r.type, enum.Enum):
+                    try:
+                        resp[p] = r.type(v)
+                    except ValueError:
+                        # Create enum dynamically
+                        resp[p] = enum.Enum("Enum", [("UNKNOWN", v)])(v)
+            except TypeError:
+                continue
+    return resp
+
+
+def use_enums(self, method):
+    """Replace values with enums when possible"""
+
+    if asyncio.iscoroutinefunction(method.func):
+
+        @functools.wraps(method)
+        async def wrapper(*args, **kwargs):
+            resp = await method(*args, **kwargs)
+            return replace_with_enums(self, resp)
+
+    else:
+
+        @functools.wraps(method)
+        def wrapper(*args, **kwargs):
+            resp = method(*args, **kwargs)
+            return replace_with_enums(self, resp)
+
+    return wrapper
 
 
 class Operation:
@@ -690,9 +926,11 @@ class Operation:
 
     def __repr__(self):
         r = self.resource
-        return f'<{self.__class__.__name__}-Resource {r.name!r} ID={r.rid}, '\
-            f'type={r.type}, range={r.range!r}, unit={r.unit!r}, ' \
-            f'mandatory={r.mandatory}, multiple={r.multiple}>'
+        return (
+            f"<{self.__class__.__name__}-Resource {r.name!r} ID={r.rid}, "
+            f"type={r.type}, range={r.range!r}, unit={r.unit!r}, "
+            f"mandatory={r.mandatory}, multiple={r.multiple}>"
+        )
 
     @property
     def ep(self):
@@ -702,21 +940,27 @@ class Operation:
     def path(self):
         oid = self.obj.oid
         if oid is None:
-            raise BadPath('Missing object ID', self.obj)
+            raise BadPath("Missing object ID", self.obj)
         iid = self.obj.iid
         if iid is None:
             iid = 0
         rid = self.resource.rid
         if rid is None:
-            raise BadPath('Missing resouce ID', self)
-        return Path(f'/{oid}/{iid}/{rid}')
+            raise BadPath("Missing resouce ID", self)
+        return Path(f"/{oid}/{iid}/{rid}")
+
+    def __getattr__(self, name):
+        meth = functools.partial(getattr(self.ep, name), self.path)
+        if name in {"read", "observe", "cancel_observe"}:
+            return use_enums(self, meth)
+        return meth
 
 
-class R(Operation, LwM2MPath):
+class R(Operation):
     pass
 
 
-class W(Operation, LwM2MPath):
+class W(Operation):
     pass
 
 
@@ -724,11 +968,11 @@ class RW(R, W):
     pass
 
 
-class E(Operation, LwM2MPath):
+class E(Operation):
     pass
 
 
-class BS_RW(Operation, LwM2MPath):
+class BS_RW(Operation):
     pass
 
 
@@ -744,13 +988,13 @@ class Resource:
 
     def __init__(self):
         """Initialization of Resource"""
-        self.name = ''
+        self.name = ""
 
     def __set_name__(self, object_def, name):
         self.name = name
 
     def __repr__(self):
-        return f'{self.name}<{self.rid}>'
+        return f"{self.name}<{self.rid}>"
 
     def __get__(self, object_instance, object_def=None):
         if object_instance is not None:
@@ -758,17 +1002,17 @@ class Resource:
         return self
 
     def __set__(self, object_instance, value):
-        raise AttributeError(f'Set attribute {self.name!r} not allowed')
+        raise AttributeError(f"Set attribute {self.name!r} not allowed")
 
 
-class ObjectDef(LwM2MPath):
+class ObjectDef:
 
     oid = None
     mandatory = True
     multiple = False
 
     def __init__(self, endpoint: Endpoint, iid=None):
-        super().__init__(endpoint)
+        self.ep = endpoint
         self.iid = iid
 
     def __init_subclass__(cls, **kwargs):
@@ -779,24 +1023,32 @@ class ObjectDef(LwM2MPath):
                 cls._rid[res.rid] = res
 
     def __repr__(self):
-        output = f'<{self.__class__.__name__} ID={self.oid}, '\
-            f'mandatory={self.mandatory}, multiple={self.multiple}>'
+        output = (
+            f"<{self.__class__.__name__} ID={self.oid}, "
+            f"mandatory={self.mandatory}, multiple={self.multiple}>"
+        )
         if self.iid is not None:
-            return output + f'[{self.iid}]'
+            return output + f"[{self.iid}]"
         return output
 
     @property
     def path(self) -> Path:
         oid = self.oid
         if oid is None:
-            raise BadPath('Missing object ID', self)
+            raise BadPath("Missing object ID", self)
         iid = self.iid
         if iid is None:
-            return Path(f'/{oid}')
-        return Path(f'/{oid}/{iid}')
+            return Path(f"/{oid}")
+        return Path(f"/{oid}/{iid}")
 
-    def __getitem__(self, key) -> 'ObjectDef':
+    def __getitem__(self, key) -> "ObjectDef":
         return self.__class__(self.ep, iid=int(key))
 
     def resource_by_id(self, rid: int):
         return self._rid.get(rid)
+
+    def __getattr__(self, name):
+        meth = functools.partial(getattr(self.ep, name), self.path)
+        if name in {"read", "observe", "cancel_observe"}:
+            return use_enums(self, meth)
+        return meth
